/*
 * Copyright 2017-2021 Thingstream AG
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/**
 * Process NMEA input from the GPS. The parser is implemented as a
 * Transport_callback, so it can directly be registered with a serial
 * transport.
 *
 * Example data from the SIM868:
 * $GNGGA,195550.000,3724.513638,N,12156.161851,W,1,11,0.97,-7.325,M,-25.528,M,,*6B
 *
 * NMEA sentences other than GGA are ignored. Gps_getQuality() can be used for
 * polling whether a fix has been obtained (quality > 0).
 */
#include <stdint.h>
#include <stdbool.h>
#include <string.h>
#include <stdlib.h>

#include "flags.h"
#include "platform.h"
#include "gps.h"

#ifdef GPS_ENABLED

/* Max length of a NMEA field (between ','); depends on GPS */
#define MAX_FIELD_LEN 16

typedef struct
{
    /* Once we have a fix, record the data here */

    struct
    {
        int32_t latitude_udeg;
        int32_t longitude_udeg;
#ifdef MANUFACTURING_TEST_MODE
        uint32_t utc;
#endif /* MANUFACTURING_TEST_MODE */
        uint8_t quality;
    } ggaFields;

    /* Parsing state */

    enum { READ_IDLE, READ_FIELD, READ_CHECKSUM } state;
    uint8_t fieldCount;  /* index of current comma-separated field */
    uint8_t fieldLen;    /* number of characters collected in field[] */
    char field[MAX_FIELD_LEN];  /* accumulate the field characters */

} gps_state_t;

static gps_state_t gps_state;


static bool processField(gps_state_t *state);
static bool parseCoord(char *ptr, uint8_t fieldLen, int32_t* coord);



/****** debug support ****/

#if defined(GPS_LOG_ENABLED) && defined(DEBUG_ENABLE)
/* Attempting to trace the gps data in the interrupt callback tends
 * to cause loss of data. Use a scheme similar to the line buffer
 * and log it in the interrupt handler, but display it in the
 * foreground via debug_dump()
 * Buffer is empty when head == tail
 * tail always points at the start of the next line to display
 */

#define LOG_SIZE 512  /* must be power of 2 */

static struct
{
    /* fields written in interrupt handler */
    volatile uint16_t head;
    volatile int16_t lastChar;
    volatile uint16_t linesFilled;

    /* fields written by foreground reader */
    uint16_t tail;
    uint16_t linesEmptied;

    char buffer[LOG_SIZE + 1];  /* extra element is always 0 */
} debug;

/* Add one character to the log ring buffer, unless it is
 * full. Store \r or \n as 0, but don't store multiple 0
 */
static void debug_log(char c)
{
    uint16_t oldHead = debug.head;
    uint16_t newHead = (oldHead + 1) & (LOG_SIZE-1);
    if (newHead == debug.tail)
        return;  /* full */

    if (c == '\r' || c == '\n')
    {
        c = 0;
        if (debug.lastChar == 0)
            return;
    }

    debug.buffer[oldHead] = c;
    debug.lastChar = c;
    debug.head = newHead;
    if (c == 0)
        debug.linesFilled += 1;
}

/* Dump any complete lines from the ring buffer */
static void debug_dump(void)
{
    while (debug.linesEmptied != debug.linesFilled)
    {
        /* We know there is a 0 just past the end of the buffer, so the string
         * starting at tail is terminated.
         */
        int tail = debug.tail;
        char *p = debug.buffer + tail;
        int len = strlen(p);

        if (tail + len >= LOG_SIZE)
        {
            /* the 0 beyond the buffer was detected, so the line wraps.
             * The first character after the wrap might be a 0, but
             * that's okay
             */
            DEBUGOUT("GPS: %s%s\n", p, debug.buffer);
            tail = 0;
            len = strlen(debug.buffer);
        }
        else
        {
            DEBUGOUT("GPS: %s\n", p);
        }

        debug.tail = (tail + len + 1) & (LOG_SIZE - 1);
        debug.linesEmptied += 1;
    }
}

#else /* GPS_LOG_ENABLED && DEBUG_ENABLE */

#define debug_log(c)
#define debug_dump()

#endif /* GPS_LOG_ENABLED && DEBUG_ENABLE */


/** @see gps.h */
void *Gps_init(bool enable)
{
    gps_state_t *gstate = &gps_state;

    /* If we are enabling GPS then also zero the ggaFields structure.
     * But if we are disabling, then do not zero to keep any results.
     */
    if (enable)
    {
        Platform_GpsEnable();
        memset(&(gstate->ggaFields), 0, sizeof(gstate->ggaFields));
        gstate->state = READ_IDLE;
    }
    else
    {
        Platform_GpsDisable();

        /* Flush anything lingering in the ring buffer */
        debug_log('\n');
        debug_dump();
    }

    return gstate;
}

/** @see gps.h */
uint8_t Gps_getQuality(void)
{
    return gps_state.ggaFields.quality;
}

/** @see gps.h */
int32_t Gps_getLatitude(void)
{
    return gps_state.ggaFields.latitude_udeg;
}

/** @see gps.h */
int32_t Gps_getLongitude(void)
{
    return gps_state.ggaFields.longitude_udeg;
}

#ifdef MANUFACTURING_TEST_MODE
uint32_t Gps_time(void)
{
    return gps_state.ggaFields.utc;
}
#endif /* MANUFACTURING_TEST_MODE */


/** @see gps.h */
void Gps_callback(void* cookie, uint8_t* data, uint16_t len)
{
    gps_state_t *gstate = cookie;

    while (len-- > 0)
    {
        char ch = *data++;
        debug_log(ch);
        if (ch == '$')
        {
            gstate->state = READ_FIELD;
            gstate->fieldLen = 0;
            gstate->fieldCount = 0;
        }
        else if (gstate->state == READ_FIELD)
        {
            if (ch == ',')
            {
                /* reached end of a field - parse it */
                if (!processField(gstate))
                    gstate->state = READ_IDLE;
            }
            else if (ch == '*')
            {
                /* reached end of last field - parse it */
                if (processField(gstate))
                    gstate->state = READ_CHECKSUM;
                else
                    gstate->state = READ_IDLE;
            }
            else
            {
                /* Add character to current field accumulator */
                uint_fast8_t len = gstate->fieldLen;
                gstate->field[len] = ch;
                if (++len >= MAX_FIELD_LEN)
                    gstate->state = READ_IDLE; /* discard line on field overflow */
                gstate->fieldLen = (uint8_t)len;
            }
        }
        else if (gstate->state == READ_CHECKSUM)
        {
            /* TODO: not yet supported */
            gstate->state = READ_IDLE;
        }
    }
}

/**
 * Parse current field.
 * This is called whenever we encounter ',' or '*'.
 * Use the information in gstate->parser to interpret the field.
 * @param gstate the gps state
 * @return true if the field was parsed successfully, false on error or when
 * encountering data not supported by the parser
 */
static bool processField(gps_state_t *gstate)
{
    bool ret = false;

    uint_fast8_t fieldLen = gstate->fieldLen;
    char *field = gstate->field;
    gstate->fieldLen = 0;  /* reset for next field */

    /* we can unconditionally increment fieldCount here since the
     * value only matters in the case of a true return.
     */
    switch (gstate->fieldCount++)
    {
        case 0: /* sentence identifier */
            if ((fieldLen == 5) && (strncmp(field + 2, "GGA", 3) == 0))
                ret = true;
            break;
        case 1: /* GPS time. */
#ifdef MANUFACTURING_TEST_MODE
            {
                char *ptr = field;
                uint32_t utc = 0;
                while (fieldLen-- > 0)
                {
                    utc *= 10;
                    utc += (*ptr++ - '0');
                }
                gstate->ggaFields.utc = utc;
            }
#endif /* MANUFACTURING_TEST_MODE */
            ret = true;
            break;
        case 2: /* lattitude */
            ret = parseCoord(field, fieldLen, &(gstate->ggaFields.latitude_udeg));
            break;
        case 3: /* N/S */
            if (fieldLen == 1)
            {
                if (field[0] == 'S')
                    gstate->ggaFields.latitude_udeg *= -1;
                ret = true;
            }
            break;
        case 4: /* longitude */
            ret = parseCoord(field, fieldLen, &(gstate->ggaFields.longitude_udeg));
            break;
        case 5: /* E/W */
            if (fieldLen == 1)
            {
                if (field[0] == 'W')
                    gstate->ggaFields.longitude_udeg *= -1;
                ret = true;
            }
            break;
        case 6: /* quality */
            if ((fieldLen == 1) && field[0] > '0')
            {
                gstate->ggaFields.quality = field[0] - '0';
                ret = true;
            }
            break;
        default: /* fall through and return false */
            break;
    }

    return ret;
}

/**
 * Parse a coordinate in the format "DDDMM.MMMMMM" and store the result as
 * microdegrees in @p coord. Uses #field and #fieldLen to access the raw
 * coordinate data.
 * @param ptr the character-sequence to parse
 * @param fieldLen the length of the character sequence
 * @param coord points to where the result should be stored
 * @return true if successfully parsed, false on error
 */
static bool parseCoord(char *ptr, uint8_t fieldLen, int32_t* coord)
{
    if (fieldLen < 8)
        return false; /* should see DDMM.MMM at least */

    char *end = ptr + fieldLen;

    /* convert "12345.67890" to 12345678
     *         "12345.6"     to 12345600
     *         "12345"       to 12345000 */
    int32_t tmp = 0;
    int scale = 0;
    while (ptr < end)
    {
        char ch = *ptr++;
        if ((ch >= '0') && (ch <= '9'))
        {
            tmp = (10 * tmp) + (ch - '0');
            scale /= 10;
        }
        else if (ch == '.')
        {
            if (scale > 0)
                return false; /* more than one '.' */
            scale = 1000;
            if (end - ptr > 3)
                end = ptr + 3; /* stop 3 digits after '.' */
        }
        else
        {
            return false;
        }
    }
    if (scale == 0)
        scale = 1000; /* no '.' seen, multiply by 1000 */
    tmp *= scale;

    *coord = (tmp / 100000 * 1000000) +  /* degrees to microdegrees */
             (tmp % 100000 * 1000 / 60); /* milliminutes to microdegrees */

    return true;
}


#ifdef DEBUG_ENABLE

/* A wrapper around Platform_wfi() that can be used while waiting for
 * gps data to become available. When gps logging is enabled, it
 * takes the opportunity to write the buffers to stdout - doing
 * it in the interrupt handler can cause data to be lost.
 */
void Gps_wfi(void)
{
    debug_dump();
    Platform_wfi();
}

#endif /* DEBUG_ENABLE */

#endif /* GPS_ENABLED */
